/*
 * Copyright (C) 2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Test file to ensure that the hand-crafted function to calculate the
 * temperature from the Pt100 resistance works as supposed and to check
 * that the error is within acceptable bounds.
 *
 */

#include <stdint.h>
#include <stdio.h>
#include <math.h>

#define ADC_MAX 0x200000UL /* 2^21 */
#define Ra 6800 /* Ohms */
#define DIVROUND(var, divisor) do { var += (divisor)/2; var /= (divisor); }while(0)

uint32_t method2(uint32_t val) {
	uint16_t Rcal = 4*Ra; /* From calibration */
	uint32_t R = (Rcal * val + ((ADC_MAX - val) >> 7)) / ((ADC_MAX - val) >> 6); /* R factor: 256 */
	uint32_t Tu;
	int32_t T;
	if(R < 100*256) {
		Tu = 4292218517UL; /* T factor: 100 * 2^34; coefficient: +2.4984e-3 * R^2 */
		Tu -= (38051 * R); /* T factor: 100 * 2^8 * 2^26; coefficient: -5.67e-6 * R^3 */
		DIVROUND(Tu, 65536); /* T factor: 100 * 2^18 */
		Tu *= R; /* T factor: 100 * 2^26 */

		DIVROUND(Tu, 8); /* T factor: 100 * 2^23 */
		Tu += 1868679872; /* T factor: 100 * 2^23; coefficient:  2.22764 * R */
		DIVROUND(Tu, 131072); /* T factor: 100 * 2^6 */
		Tu *= R; /* T factor: 100 * 2^14 */
		T = Tu;
		T -= 396620595; /* T factor: 100 * 2^14; coefficient: -242.078 */
		DIVROUND(T, 16384); /* T factor: 100 */
	} else {
		// R MAX: 91752
		Tu = 35942UL * R ; /* T factor: 100 * 2^38 * 2^8; coefficient: 1.30755e-09 * R^4 */

		Tu += 549920477UL; /* T factor: 100 * 2^46; coefficient: 7.81484e-08 * R^3 */
		DIVROUND(Tu, 131072UL); /* T factor: 100 * 2^29 */
		Tu *= R; /* T factor: 100 * 2^37 */

		DIVROUND(Tu, 8); /* T factur: 100 * 2^34 */
		Tu += 1504836281; /* T factor: 100 * 2^34; coefficient: 0.00087593 * R^2 */
		DIVROUND(Tu, 131072UL); /* T factor: 100 * 2^17 */
		Tu *= R; /* T factor: 100 * 2^25 */

		DIVROUND(Tu, 4); /* T factor: 100 * 2^23 */
		Tu += 1992806105; /* T factor: 100 * 2^23; coefficient: 2.37561 * R */
		DIVROUND(Tu, 65536UL); /* T factor: 100 * 2^7 */
		Tu *= R; /* T factor: 100 * 2^15 */

		Tu /= 2; /* T factor: 100 * 2^14 */
		T = Tu;
		T -= 403909837;
		DIVROUND(T, 16384UL); /* T factor: 100 */
	}
	return T;
}

uint32_t method_orig_neg(uint32_t val) {
uint32_t R = (2 * Ra * val + ((ADC_MAX - val) >> 7)/2) / ((ADC_MAX - val) >> 7); /* R factor: 256 */
//	uint32_t R = (4 * Ra * val + ((ADC_MAX - val) >> 7)) / ((ADC_MAX - val) >> 6); /*  R factor: 256 */
  int32_t T;
  T = (1452 * R); /* T factor: 65536 * 1 000 000 */
	T = -T;
  T += 163735142; /* T factor: 65536 * 1 000 000 */
	T += 65536/2;
  T /= 65536; /* T factor: 1 000 000 */
  T *= R; /* T factor: 256 * 1 000 000 */

	T += 5;
	T /= 10; /* T factor: 256 * 100 000 */

  T += 57027584; /* T factor: 256 * 100 000 */
	T += 2048;
  T /= 4096; /* T factor: 100 000 / 16 */
  T *= R; /* T factor: 16 * 100 000 */
	T += 8;
  T /= 16;
  T -= 24207800; /* T factor:* 100 000 */
	T += 500;
  T /= 1000;
	return T;
}

int main(int argc, char* argv[]) {
	uint32_t val; /* = atol(argv[1]); */
	for(val = 20000; val < 105000; val++) {
		double r = 1.0 * Ra * val / (ADC_MAX - val);
		//printf("R %f\n", r);
		double Tn = -5.67e-6 * r*r*r + 2.4984e-3 * r*r + 2.22764 * r - 242.078;
		double Tp = 3383.81 - sqrt(1.31818*(10000000) - 17316*r);
		double Tf = r < 100 ? Tn : Tp;
		int32_t T = method2(val);
		int32_t Torig = method_orig_neg(val);
		printf("%u %f %f %f %f\n", val, T/100.0, Tf, Tf - T/100.0, Tf-Torig/100.0);
	}
	return 0;
}

