// coding: utf-8
// ----------------------------------------------------------------------------
/*
 * Main part:
 * Copyright (c) 2007 Roboterclub Aachen e.V.
 *  All rights reserved.
 *
 * pow2:
 * Copyright (c) 2007,2008 Kip Macy kmacy@freebsd.org
 * All rights reserved.
 *
 * Other extensions:
 * Copyright (c) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// ----------------------------------------------------------------------------

#ifndef	UTILS_H
#define	UTILS_H

#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>
#include <stdint.h>
#include <stddef.h>

#include "config.h"

#ifndef	TRUE
	#define	TRUE	(1==1)
#elif !TRUE
	#error	Faulty definition for TRUE
#endif

#ifndef FALSE
	#define	FALSE	(1!=1)
#elif FALSE
	#error Faulty definition for FALSE
#endif

#ifndef NULL
	#define NULL ((void*)0)		//!< Nullzeiger
#endif


#ifndef MIN
	#define MIN(a,b) ((a)<(b)? (a) : (b))
#endif

#ifndef MAX
	#define MAX(a,b) ((a)>(b)? (a) : (b))
#endif

/*
 * Comparation primitives to enforce an 8-bit comparation instead
 * of the default 16-bit:
 */
#define EQ8(a, b) ((uint8_t)(a) == (uint8_t)(b))
#define NE8(a, b) ((uint8_t)(a) != (uint8_t)(b))
#define LT8(a, b) ((uint8_t)(a) < (uint8_t)(b))
#define LE8(a, b) ((uint8_t)(a) <= (uint8_t)(b))
#define GT8(a, b) ((uint8_t)(a) > (uint8_t)(b))
#define GE8(a, b) ((uint8_t)(a) >= (uint8_t)(b))


/* TODO: Check if this is smaller: */
/*
 * #define	LO16(x)		(((uint32_t8) ((uint32_t)(x))).lo)
 * #define	HI16(x)		(((uint32_t8) ((uint32_t)(x))).hi)
 */ 

#define	LO8(x)		((uint8_t) ((x) & 0xff))
#define	HI8(x)	((uint8_t) ((x) >> 8))
#define LO16(x)		((uint16_t) ((x) & 0xffff))
#define HI16(x)    ((uint16_t) ((x) >> 16))

typedef union {
	uint8_t b[4];
	struct {
		uint8_t b1;		// lsb
		uint8_t b2;
		uint8_t b3;
		uint8_t b4;		// msb
	};
	struct {
		uint16_t lo;
		uint16_t hi;
	};
	uint32_t val;
} uint32_t8;

typedef union {
	uint8_t b[2];
	struct {
		uint8_t lo;
		uint8_t hi;
	};
	uint16_t val;
} uint16_t8;

/* Critical section macros for better readability: */

#if __AVR_LIBC_VERSION__ >= 10600 && !defined (__cplusplus)

	#include <util/atomic.h>
	
	#define	ENTER_CRITICAL_SECTION		ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
	#define	LEAVE_CRITICAL_SECTION		}

#else

	#define	ENTER_CRITICAL_SECTION		do { unsigned char sreg_ = SREG; cli();
	#define	LEAVE_CRITICAL_SECTION		SREG = sreg_; } while (0);

#endif

/* Atomic byte exchange: */

static inline uint8_t read_and_replace_atomic(volatile uint8_t *data, uint8_t new_data)
{
	uint8_t old_data;
	
	ENTER_CRITICAL_SECTION
	
	// Daten tauschen
	old_data = *data;
	*data = new_data;
	
	LEAVE_CRITICAL_SECTION
	
	return old_data;
}

/* Volatile access to variables: */

#define	vu8(x)	(*(volatile  uint8_t*)&(x))
#define	vs8(x)	(*(volatile   int8_t*)&(x))
#define	vu16(x)	(*(volatile uint16_t*)&(x))
#define	vs16(x)	(*(volatile  int16_t*)&(x))
#define	vu32(x)	(*(volatile uint32_t*)&(x))
#define	vs32(x)	(*(volatile  int32_t*)&(x))


/* More compiler magic. Force GCC to use a local pointer: */
#define FIX_POINTER(_ptr) __asm__ __volatile__("" : "=b" (_ptr) : "0" (_ptr))

/**
 * \ingroup	utils_h
 * \name	Port-Makros
 *
 * Die Makros RESET(), SET(), SET_OUTPUT(), SET_INPUT() und IS_SET()
 * beziehen sich immer auf ein bestimmtes Bit eines Ports und helfen somit
 * den Code sehr portabel zu gestalten.
 *
 * Beispiel:
 * \code
 * #define LED   D,5		// PORTD, Pin 5
 *
 * SET_OUTPUT(LED);		// Pin als Ausgang schalten (wird z.B. zu DDRD |= (1<<5);)
 * 
 * SET(LED);				// LED aktivieren
 * \endcode
 *
 * oder auch:
 *
 * \code
 * #define SCHALTER   B,1		// PORTB, Pin 1
 * 
 * SET_INPUT_WITH_PULLUP(SCHALTER);
 * 
 * if (IS_SET(SCHALTER)) {
 * 		...
 * }
 * \endcode
 * 
 * Somit muss nur ein Define geändert werden sobald ein anderer Pin verwendet
 * werden soll. Außerdem muss nicht immer noch ein extra Define für den 
 * entsprechenden Port angelegt werden wie es bisher immer der Fall war.
 */

/* Warum hier zum Teil so seltsame Konstrukte notwendig sind wird zum Beispiel
 * in http://www.mikrocontroller.net/forum/read-1-324854.html#324980 erklärt.
 */
#define	PORT(x)			_port2(x)
#define	DDR(x)			_ddr2(x)
#define	PIN(x)			_pin2(x)
#define	REG(x)			_reg(x)
#define	PIN_NUM(x)		_pin_num(x)

#define	RESET(x)		RESET2(x)
#define	SET(x)			SET2(x)
#define	TOGGLE(x)		TOGGLE2(x)
#define	SET_OUTPUT(x)	SET_OUTPUT2(x)
#define	SET_INPUT(x)	SET_INPUT2(x)
#define	SET_PULLUP(x)	SET2(x)
#define	IS_SET(x)		IS_SET2(x)

#define	SET_INPUT_WITH_PULLUP(x)	SET_INPUT_WITH_PULLUP2(x)

#define	_port2(x)	PORT ## x
#define	_ddr2(x)	DDR ## x
#define	_pin2(x)	PIN ## x

#define	_reg(x,y)		x
#define	_pin_num(x,y)	y

#define	RESET2(x,y)		PORT(x) &= ~(1<<y)
#define	SET2(x,y)		PORT(x) |= (1<<y)
#define	TOGGLE2(x,y)	PORT(x) ^= (1<<y)

#define	SET_OUTPUT2(x,y)	DDR(x) |= (1<<y)
#define	SET_INPUT2(x,y)		DDR(x) &= ~(1<<y)
#define	SET_INPUT_WITH_PULLUP2(x,y)	SET_INPUT2(x,y);SET2(x,y)

#define	IS_SET2(x,y)	((PIN(x) & (1<<y)) != 0)


// ----------------------------------------------------------------------------
/**
 * \ingroup	utils_h
 * \name	weitere nuetzliche Makros
 */
//@{
#if defined(__DOXYGEN__)

#define	_bit_is_set(pin, bit)	//!< Überprüft ob das entsprechende Bit gesetzt ist
#define	_bit_is_clear(pin, bit)	//!< Überprüft ob das entsprechende Bit gelöscht ist

#define	STRING(x)		//!< erstellt einen String aus dem übergebenen Wert

#else /* !DOXYGEN */

#define	_bit_is_set(pin, bit)	(pin & (1<<bit))
#define	_bit_is_clear(pin, bit)	(!(pin & (1<<bit)))

#define	STRING(x)	_STRING(x)
#define	_STRING(x)	#x

#endif /* DOXYGEN */
//@}

// ----------------------------------------------------------------------------
/**
 * \ingroup	utils_h
 * \brief	Dreht die beiden Nibble in einem Byte um.
 *
 * \param	x	Byte das verarbeitet werden soll.
 */
static inline uint8_t swap (uint8_t x)
{
	if (__builtin_constant_p(x))
		x = (x << 4) | (x >> 4);
	else
		asm volatile ("swap %0" : "=r" (x) : "0" (x));
	
	return x;
}

// ----------------------------------------------------------------------------
/**
 * \ingroup	utils_h
 * \brief	
 */
#if defined(DEBUG_LEVEL) && DEBUG_LEVEL
	#include <stdio.h>
	#define	DEBUG_PRINT(level, s, ...)	do { if(level <= DEBUG_LEVEL) { \
			static prog_char __s[] = (s); \
				printf_P(__s, ## __VA_ARGS__); } } while (0)
	#define	DEBUG_PUT(level, s)	do { if(level <= DEBUG_LEVEL) { \
				puts_P(PSTR(s)); } } while (0)
#else
	#define	DEBUG_PRINT(level, s, ...)
	#define	DEBUG_PUT(level, s)
#endif

/* Non-bootloader informational output: */
#ifndef IS_BOOTLOADER
# define INFO_PRINT(s) puts_P(PSTR(s))
#else
# define INFO_PRINT(s) 
#endif

// ----------------------------------------------------------------------------
/**
 * \ingroup	utils_h
 * \brief	Zählt die Anzahl der gesetzten Bits in einem Byte
 *
 * \param	n	Byte das verarbeitet werden soll.
 * \return	Anzahl der gesetzten Bits (0..8)
 *
 * \see		http://infolab.stanford.edu/~manku/bitcount/bitcount.html
 */
static inline uint8_t bit_count8(uint8_t n)
{
	n = ((n >> 1) & 0x55) + (n & 0x55);
	n = ((n >> 2) & 0x33) + (n & 0x33);
	n = ((n >> 4) + n) & 0xf;
	
	return n;
}

#define MASK_01010101 (((uint32_t)(-1))/3)
#define MASK_00110011 (((uint32_t)(-1))/5)
#define MASK_00001111 (((uint32_t)(-1))/17)

// ----------------------------------------------------------------------------
/**
 * \ingroup	utils_h
 * \brief	Zählt die Anzahl der gesetzten Bits in einem Byte
 * 
 * \param	n	Wert der verarbeitet werden soll.
 * \return	Anzahl der gesetzten Bits (0..32)
 */
static inline uint8_t bit_count32(uint32_t n)
{
	n = (n & MASK_01010101) + ((n >> 1) & MASK_01010101);
	n = (n & MASK_00110011) + ((n >> 2) & MASK_00110011);
	n = (n & MASK_00001111) + ((n >> 4) & MASK_00001111);
	
	return n % 255 ;
}

/* host to network conversion (for static expressions): */
#define NTOHL(x) HTONL(x)
#define HTONL(x) \
        ((((x) & 0xff000000) >> 24) | \
         (((x) & 0x00ff0000) >>  8) | \
         (((x) & 0x0000ff00) <<  8) | \
         (((x) & 0x000000ff) << 24))

/* HTONxf for dynamic expressions: 
 * These evaluate to much shorter code then HTONx.
 */

#define HTONSf(x) HTONS_((uint16_t8) ((uint16_t)(x)))
static inline uint16_t HTONS_(uint16_t8 x) {
	uint32_t8 ret;
	ret.hi = x.lo;
	ret.lo = x.hi;
	return ret.val;
}

#define HTONLf(x) HTONL_((uint32_t8) ((uint32_t)(x)))
static inline uint32_t HTONL_(uint32_t8 x) {
	uint32_t8 ret;
	ret.b1 = x.b4;
	ret.b2 = x.b3;
	ret.b3 = x.b2;
	ret.b4 = x.b1;
	return ret.val;
}

/* CRC8 generation: */
uint8_t crc8 (uint8_t crc, uint8_t data);
uint8_t gen_crc8(uint8_t *data, uint16_t len);

/* unsigned divisions: */
typedef struct
{
	uint32_t quot;
	uint32_t rem;
} ludiv_t;

extern ludiv_t ludiv (uint32_t, uint32_t) __asm__("__udivmodsi4");


/* Arithmetics on accumulator pointers: */
void inc32(uint32_t *accumulator);
void add32_16(uint32_t *accumulator, uint16_t value);
void sub32_16(uint32_t *accumulator, uint16_t value);
void add16_16(uint16_t *accumulator, uint16_t value);
void sub16_16(uint16_t *accumulator, uint16_t value);

/* Special calls to modify longer types: */
void clear32(uint32_t *num); /* Set *num to 0 */

/* AVR gcc attributes */
#define noinline __attribute__ ((noinline))
#define noreturn __attribute__ ((noreturn))
#define inlined __attribute__ ((always_inline))

/* Device reboot methods: */
void reboot_device(void);
void reboot_to_bootloader(void);

/* Number parse and print methods: */
struct parse_ret16 {
	char *next;
	uint16_t val;
};
/*
 * parse_num16:
 * Parse the digit string at start to the given base until non-digit chars
 * occur. A pointer to the first invalid char and the parsed value is
 * returned.
 */
struct parse_ret16 parse_num16(const char* start, uint8_t base);

/*
 * parse_num32:
 * Parse a number starting at "str", represented in the given "base", put
 * it into "dst", if successful.
 * If "str" is NULL, or no literal could be parsed, NULL is returned.
 * Else, the return value it points to the first non-digit character in "str".
 * Digits are 0-9, followed by A-Z / a-z ('a'=='A') for bases > 10.
 */
const char* parse_num32(const char* str, uint32_t* dst, uint8_t base);

/*
 * print_num16/32:
 * Put the supplied number in the supplied base into the string at start.
 * The string is null-terminated, but returned is a pointer to this termina-
 * ting \0 to ease concatenation.
 */
char* print_num32(char* start, uint32_t num, uint8_t base);
char* print_num16(char* start, uint16_t num, uint8_t base);

/*
 * stpcpy_P:
 * Copy the string str, residing in the PROGMEM, to start and return
 * a pointer to the NULL-termination character put after the string.
 */
char* stpcpy_P(char* start, const char* str);

/*
 * stpcpy:
 * Copy the string "str" to "start" and return a pointer to the NULL-
 * termination character put after the string.
 */
char* stpcpy(char* start, const char* str);

/* Number / 2-exponent conversion: */
#define SNUM2EXP(num) \
	( \
		((num) == 1)?  0 : \
		((num) == 2)?  1 : \
		((num) == 4)?  2 : \
		((num) == 8)?  3 : \
		((num) == 16)?  4 : \
		((num) == 32)?  5 : \
		((num) == 64)?  6 : \
		((num) == 128)?  7 : \
		((num) == 256)?  8 : \
		((num) == 512)?  9 : \
		((num) == 1024)? 10 : \
		((num) == 2048)? 11 : \
		((num) == 4096)? 12 : \
		((num) == 8192)? 13 : \
		((num) == 16384)? 14 : \
		((num) == 32768)? 15 : \
		((num) == 65536)? 16 : \
		0xFFFFFFFFL \
	) 

/* Check if the given value is a power of two: */
#define pow2(x)       ((((x)-1)&(x))==0)

/* Evaluates to the offset of the element "m" in a struct "t": */
#ifndef offsetof
  #define offsetof(t,m) (uint16_t)((&((t *)0L)->m))
#endif

/* Forward declaration, since strings.h doesn't exist. */
void bzero(void *b, size_t len);

#endif	// UTILS_H
