/*
 * Copyright (C) 2011-2012 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <stdint.h>
#include "config.h"
#include "avrutils.h"
#include "lcd.h"

#ifndef F_CPU
#error Please define the CPU frequency F_CPU
#endif

/* Read flag: */
#define LCD_BUSY_FLAG 0x80

#ifndef LCD_SLOWNESS_MULTIPLIER
#define LCD_SLOWNESS_MULTIPLIER 1
#endif

/* Only sleep if clock is too fast for 220 ns E length: */
#if LCD_SLOWNESS_MULTIPLIER == 1
	#if F_CPU <= 4545454/2
		#define lcd_pulse_wait() do {/* no nop */}while(0)
	#elif F_CPU <= 9090909/2
		#define lcd_pulse_wait() asm("nop\n\t")
	#elif F_CPU <= 13636363/2
		#define lcd_pulse_wait() asm("nop\n\tnop\n\t")
	#elif F_CPU <= 18181818/2
		#define lcd_pulse_wait() asm("nop\n\tnop\n\tnop\n\t")
	#elif F_CPU <= 22727272/2
		#define lcd_pulse_wait() asm("nop\n\tnop\n\tnop\n\ŧnop\n\t")
	#elif F_CPU <= 27272727/2
		#define lcd_pulse_wait() asm("nop\n\tnop\n\tnop\n\ŧnop\n\tnop\n\t")
	#elif F_CPU <= 31818181/2
		#define lcd_pulse_wait() asm("nop\n\tnop\n\tnop\n\ŧnop\n\tnop\n\tnop\n\t")
	#else
		#error Please extend, your processor is too fast
	#endif
#else
	#define lcd_pulse_wait() _delay_us(LCD_SLOWNESS_MULTIPLIER * .220);
#endif

#define lcd_pulse_enable() \
	do {SET(LCD_PIN_E); lcd_pulse_wait(); RESET(LCD_PIN_E);} while(0)

void lcd_exec(uint8_t val) {
	RESET(LCD_PIN_RW);
	#if LCD_MODE_4BIT
		LCD_DATA_PORT &= ~0x0F << LCD_DATA_SHIFT;
		LCD_DATA_PORT |= (val >> 4) << LCD_DATA_SHIFT;
		lcd_pulse_enable();

		LCD_DATA_PORT &= ~0x0F << LCD_DATA_SHIFT;
		LCD_DATA_PORT |= (val & 0x0F) << LCD_DATA_SHIFT;
		lcd_pulse_enable();
	#else
		LCD_DATA_PORT = val;
		lcd_pulse_enable();
	#endif
	_delay_us(LCD_SLOWNESS_MULTIPLIER * 46);
}

void lcd_exec_cmd(uint8_t cmd) {
	RESET(LCD_PIN_RS);
	lcd_exec(cmd);
	SET(LCD_PIN_RS);
}
#define lcd_send_data(data) lcd_exec(data)

void lcd_init(void) {
	RESET(LCD_PIN_RW);
	RESET(LCD_PIN_E);
#if !LCD_OWN_INIT
	SET_OUTPUT(LCD_PIN_E);
	SET_OUTPUT(LCD_PIN_RW);
	SET_OUTPUT(LCD_PIN_RS);
	#if LCD_MODE_4BIT
		LCD_DATA_DDR |= 0x0F << LCD_DATA_SHIFT;
	#else
		LCD_DATA_DDR = 0xFF;
	#endif
#endif
	/*
	 * Initialisation sequence using a homing sequence that
	 * brings the HD44780 display into a known state, regardless
	 * of its previous state:
	 */

	/* Wait for the display to get ready: */
	_delay_ms(LCD_SLOWNESS_MULTIPLIER * 15);

	/* First iteration, switch to 8-bit interface: */
	RESET(LCD_PIN_RW);
	RESET(LCD_PIN_RS);
	#if LCD_MODE_4BIT
		LCD_DATA_PORT &= ~(0x0F << LCD_DATA_SHIFT);
		LCD_DATA_PORT |= ((CMD_SET_FUNCTION | LCD_FUNC_8BIT) >> 4)
			<< LCD_DATA_SHIFT;
	#else
		LCD_DATA_PORT = (CMD_SET_FUNCTION | LCD_FUNC_8BIT);
	#endif
	uint8_t i = 3;
	do {
		lcd_pulse_enable();
		if(i == 3)
			_delay_ms(LCD_SLOWNESS_MULTIPLIER * 5);
		else
			_delay_us(LCD_SLOWNESS_MULTIPLIER * 100);
	} while(--i);

	/* LCD is now initialised to: 8-bit, 5x7, 1-line, all off, +1 increment, no shift */

	#if LCD_MODE_4BIT
		/* Switch to 4-bit mode: */
		LCD_DATA_PORT = ((CMD_SET_FUNCTION | LCD_FUNC_4BIT) >> 4) << LCD_DATA_SHIFT;
		lcd_pulse_enable();
		_delay_us(LCD_SLOWNESS_MULTIPLIER * 42);
	#endif

	/* Set display capabilities. Lines and font may only be once, here! */
	/* XXX: This only works for 5x7! */
	lcd_exec(CMD_SET_FUNCTION |
		LCD_FUNC_5x7px | (LCD_LINES == 2 ? LCD_FUNC_2_LINE : LCD_FUNC_1_LINE)
		| (LCD_MODE_4BIT ? LCD_FUNC_4BIT : LCD_FUNC_8BIT));
	_delay_us(LCD_SLOWNESS_MULTIPLIER * 42);
	/* Bring the display into a sane state and switch to data mode: */
	lcd_clear();
	lcd_home();
}

void lcd_clear() {
	lcd_exec_cmd(CMD_DISPLAY_CLEAR);
	_delay_ms(LCD_SLOWNESS_MULTIPLIER * 1.64);
}

void lcd_home() {
	lcd_exec_cmd(CMD_RETURN_HOME);
	_delay_ms(LCD_SLOWNESS_MULTIPLIER * 1.64);
}

void lcd_puts(char* s) {
	char c;
	while((c = *(s++))) lcd_putc(c);
}

void lcd_putS(const char* s) {
	char c;
	while((c = pgm_read_byte(s++))) lcd_putc(c);
}

void lcd_loadchar(uint8_t charno, const uint8_t* content_P) {
	/* XXX: This only works for 5x7! */
	lcd_exec_cmd(CMD_SET_CGRAM_ADDR | ((charno) << 3));
	uint8_t i = 8;
	do {
		lcd_putc(pgm_read_byte(content_P++));
	} while(--i);
}
