/*
 * Copyright (C) 2011-2012 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>

/* Write commands: */
#define CMD_DISPLAY_CLEAR 0x01
#define CMD_RETURN_HOME 0x02
#define CMD_SET_ENTRY_MODE 0x04
#define CMD_SET_DISPLAY 0x08
#define CMD_SET_SHIFT 0x10
#define CMD_SET_FUNCTION 0x20
#define CMD_SET_CGRAM_ADDR 0x40
#define CMD_SET_DDRAM_ADDR 0x80


/* Entry mode: */
#define LCD_CURSER_MOVE_INCREASE 0x2
#define LCD_CURSER_MOVE_DECREASE 0x0
#define LCD_DISPLAY_SHIFTED 0x1
#define LCD_DISPLAY_UNSHIFTED 0x0

/* Display modes: */
#define LCD_BLINKING 0x1
#define LCD_NOT_BLINKING 0x0
#define LCD_CURSOR_ON 0x2
#define LCD_CURSOR_OFF 0x0
#define LCD_DISPLAY_ON 0x4
#define LCD_DISPLAY_OFF 0x0

/* Shift modes: */
#define LCD_SHIFT_DISPLAY 0x4
#define LCD_MOVE_CURSOR 0x0
#define LCD_SHIFT_RIGHT 0x8
#define LCD_SHIFT_LEFT 0x0

/* Function modes: */
#define LCD_FUNC_5x10px 0x4
#define LCD_FUNC_5x7px 0x0
#define LCD_FUNC_2_LINE 0x8
#define LCD_FUNC_1_LINE 0x0
#define LCD_FUNC_8BIT 0x10
#define LCD_FUNC_4BIT 0

/* Useful defaults: */
#ifndef LCD_LINE_LEN
#  define LCD_LINE_LEN 0x40 /* Next line at 0x40, not 0x10 for 2x16 LCDs */
#endif

/*
 * lcd_init:
 * Initialise the LCD to a known state (cursor off, display on, blinking off).
 * No shift, cursor moves right.
 */
void lcd_init(void);

/*
 * lcd_clear:
 * Clears the LCD.
 */
void lcd_clear(void);

/*
 * lcd_home:
 * Move the cursor to the home position (left top).
 */
void lcd_home(void);

/*
 * lcd_putc:
 * Puts the character into the display.
 */
#define lcd_putc(d) lcd_exec((uint8_t)d)

/*
 * lcd_set_shift:
 * Set the display shift/cursor move and shift right/left bits.
 */
#define lcd_set_shift(shift_move, right_left) \
	lcd_exec_cmd(CMD_SET_SHIFT | (shift_move) | (right_left))

/*
 * lcd_set_display:
 * Set the display on/off, cursor on/off, blinking on/off bits.
 */
#define lcd_set_display(display, cursor, blinking) \
	lcd_exec_cmd(CMD_SET_DISPLAY | (display) | (cursor) | (blinking))

/*
 * lcd_set_entry_mode:
 * Set the cursor move direction (increase/decrease) and display shift mode 
 * (on/off).
 */
#define lcd_set_entry_mode(cursor_dir, shift_mode) \
	lcd_exec_cmd(CMD_SET_ENTRY_MODE | (cursor_dir) | (shift_mode))

/*
 * lcd_puts:
 * Puts a whole string into the display.
 */
void lcd_puts(char* s);

/*
 * lcd_putS:
 * Puts a whole PROGMEM string into the display.
 */
void lcd_putS(const char* s);

/*
 * lcd_gotoxy:
 * Moves the cursor to the specified position (left top is 0,0).
 * XXX: Only works for 2x16 displays for now?
 */
#define lcd_gotoxy( x, y) do { uint8_t pos; \
	  pos = (x) & 0x3F; if(y) pos |= LCD_LINE_LEN; \
		lcd_exec_cmd(CMD_SET_DDRAM_ADDR | pos); \
	} while(0)


/*
 * lcd_loadchar:
 * Load the character definition at "content_P" (resides in the flash)
 * into the display, using special character number "charno".
 * ATTENTION: You have to you lcd_gotoxy() afterwards to continue drawing 
 * characters on the display!
 */
void lcd_loadchar(uint8_t charno, const uint8_t* content_P);

/* Only used internally: */
void lcd_exec(uint8_t);
void lcd_exec_cmd(uint8_t cmd);
