/*
 * Copyright (C) 2011-2012 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Fuses:
 * -U lfuse:w:0xe2:m -U hfuse:w:0xd4:m -U efuse:w:0xff:m
 * (8 MHz internal RC osc., 4.3V BOD, preserve EEPROM, enable ISP)
 */

#include "avrutils.h"
#include "config.h"
#include "lcd.h"
#include "uart_soft.h"

#include <avr/wdt.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <stdio.h>

#ifndef F_CPU
#error Please define the CPU frequency F_CPU
#endif

#define VERSION "v1.2"

static uint32_t read_adc(void);
static int32_t correct_pt(uint32_t val);
static void print_temp(uint8_t digits);
static void check_cal(void);
static uint16_t get_Rcal(void);
static uint8_t spi_read8(void);

char lcd_buf[16]; /* +\0 + extra char for Ra calibration output */
int32_t temp;
uint8_t in_main_loop = 0;

#define EEP_RA_ADDR 2
#define Rcal_multiplicator 4
#define CAL_PIN A,7
uint16_t Rcal;

#define ADCsamples 16

#define DEGREE "\01"
#define HEART '\02'
#define DEGREEC '\04'
const uint8_t char_degree[] PROGMEM = {7, 5, 7, 0, 0, 0, 0, 0};
const uint8_t char_heart[] PROGMEM = {0, 0, 10, 21, 17, 10, 4, 0};
const uint8_t char_heart2[] PROGMEM = {0, 0, 10, 31, 31, 14, 4, 0};
const uint8_t char_degC[] PROGMEM = {24, 26, 5, 4, 4, 4, 5, 2};

//int main(void) __attribute__((noreturn));
int main(void) {
	/*	Forcefully disable the watchdog. Else we end in an endless reset loop. */
#ifdef MCUCSR
	MCUCSR = 0;
#elif defined(MCUSR)
	MCUSR = 0;
#endif
	wdt_disable();

	/* Bring all outputs into a known state: */
	PORTA = 0;
	PORTB = 0;

	/* Set outputs: */
	DDRA=_BV(0) | _BV(1) | _BV(2) | _BV(3) | /* LCD */
		_BV(4); /* SPI-SCK */
	DDRB = _BV(0) | _BV(1) | _BV(2); /* LCD */

	lcd_init();


	lcd_set_entry_mode(LCD_CURSER_MOVE_INCREASE, LCD_DISPLAY_UNSHIFTED);
	lcd_set_display(LCD_NOT_BLINKING, LCD_CURSOR_OFF, LCD_DISPLAY_ON);
	lcd_set_shift(LCD_MOVE_CURSOR, LCD_SHIFT_RIGHT);
	lcd_loadchar(1, char_degree);
	lcd_loadchar(2, char_heart);
	lcd_loadchar(3, char_heart2);
	lcd_loadchar(4, char_degC);

	/* Set USI to 0,0-SPI mode (sample DI on rising CLK edge): */
	USICR = (0 << USIWM1) | ( 1 << USIWM0) | (1 << USICS1) | (0 << USICS0) |
		(1 << USICLK);

	lcd_gotoxy(0,0);
	lcd_home();
	lcd_putS(PSTR("Pt100 DS"));
	lcd_gotoxy(0, 1);
	lcd_putS(PSTR("2011-12"));
	_delay_ms(3000);

	/* Read old calibration data and check if an update is requested: */
	Rcal = eeprom_read_word((void*) EEP_RA_ADDR);
	check_cal();

  /* Initialize UART after CAL check since CAL uses the same pin. */
  uart_soft_init();
  uart_puts_P(PSTR("Pt100 Thermometer " VERSION " \n  (C) 2011 Dennis Schneider\n"));

	lcd_clear();
	lcd_home();

  // Ignore first garbage measurement:
  read_adc();

	in_main_loop = 1;
	for(;;) {
		uint32_t raw_val = read_adc();
		lcd_gotoxy(0,1); lcd_putc(' ');
		temp = correct_pt(raw_val);
		strcpy_P(lcd_buf, PSTR("        "));
		print_temp(4);

		/* Insert decimal separator: */
		lcd_buf[6] = lcd_buf[5];
		lcd_buf[5] = lcd_buf[4];
		lcd_buf[4] = ',';
		/* Correct leading zero and zero digits after the decimal separator: */
		for(uint8_t i = 0; i < 4; i++)
			if(lcd_buf[3 + i] == ' ') lcd_buf[3 + i] = '0';

		lcd_buf[7] = DEGREEC;
		lcd_gotoxy(0,0); lcd_puts(lcd_buf);

    lcd_buf[4] = '.';
    {
      uint8_t i = 0;
      while(lcd_buf[i] == ' ') ++i;
      --i;
      while(++i < 7)
        uart_send(lcd_buf[i]);
      uart_puts_P(PSTR(" DEG C"));
    }
	}
}

static uint8_t spi_read8(void) {
	USISR = _BV(USIOIF);
	uint8_t usicr_tc_strobe = USICR | _BV(USITC);
	do {
		USICR = usicr_tc_strobe;
	} while((uint8_t)(USISR & _BV(USIOIF)) == 0);
	return USIDR;
}

static uint32_t read_adc(void) {
	uint8_t show_heart = in_main_loop;
	static uint8_t heart_info = 0;
	uint8_t sample_count = ADCsamples;
	uint32_t accu = 0;

	do {

		if(show_heart) {
			uint8_t beat = 0;
			if(heart_info & 0x80) {
				if(((++heart_info) & 0x7F) > 1) {
					beat = 1;
					heart_info = 0x00;
				}
			} else if(++heart_info > 8) {
				beat = 1;
				heart_info = 0x80;
			}
			lcd_gotoxy(7,1); lcd_putc(beat ? HEART + 1 : HEART);
		}

		/* Read the first (ready) bit, beginning with a rising CLK edge: */
		PORTA |= _BV(4);

			/* Since we idle here most of the time, reduce the clock: */
			CLKPR = _BV(CLKPCE);
			CLKPR = _BV(CLKPS3);

		while(PINA & _BV(6)); /* Wait for /RDY, until a conversion finished */

			/* Resume fast speed: */
			CLKPR = _BV(CLKPCE);
			CLKPR = 0;

		/* Falling edge of clock emulation: */
		PORTA &= ~_BV(4);

		/* Now, the 24 bits of the conversion follow: */
		uint32_t8 raw;
		raw.b[3] = 0;
		raw.b[2] = spi_read8();
		raw.b[1] = spi_read8();
		raw.b[0] = spi_read8();
		accu += raw.val;

	} while(--sample_count);

	accu += ADCsamples/2;
	return accu/ADCsamples;

}

static int32_t correct_pt(uint32_t val) {
	/*
	 * Notice: Since "val" contains 22 bits + OH/OL in 2-complement's notation
	 * and we only expect positive values, the MSB may be left untouched.
	 *
	 * Some calculations:
	 * R = Ra * (val/(2^(n-1) - val)) with Ra = 6.80kOhm, n = 22 w/ MCP3551
	 * R in [60.000 0, 350.000] range -> val < 105 000
	 *      interesting: [88 150] Ohm, fit:
	 *   R' = (2*Ra * val) / ((2^(n-1) - val) / 128) = 256*R
	 *
	 * T < 0:
	 *   From wikipedia:
	 *   T = -5.67e-6 * R^3 + 2.4984e-3 * R^2 + 2.22764 * R - 242.078
	 * T > 0:
	 *   via gnuplot:
	 *   T = 1.15033e-06 * R^3 + 0.000560899 * R^2 + 2.4147 * R - 248.251
	 *   T = 1.30755e-09 * R^4 + 7.81484e-08 * R^3 + 0.00087593 * R^2 + 2.37561 * R - 246.527
	 *   (direct solution (3383.81-sqrt(1.31818*(10**7) - 17316*x)))
	 *    using values of:
	 *     R = 100 * (1 + 3.9083e-3 * T - 5.775e-7 * T^2)
	 *
	 *
	 * Error analysis (integer version vs. double version):
	 * -100 to   0 °C: -0.03  -- +0.005 °C
	 *    0 to 100 °C: -0.03  -- +0.005 °C
	 *  ... linear error growth up to:
	 *         600 °C: -0.06  -- +0.014 °C
	 *  ... then steer upwards ...
	 *         800 °C  +0.019 -- +0.087 °C (error becomes only positive)
	 *
	 */
#define ADC_MAX 0x200000UL /* 2^21 */
#define Ra 6800 /* Ohms */
#define DIVROUND(var, divisor) do{var += (divisor)/2; var /= (divisor);}while(0)
	/* Calculate R as Ra (6800 Ohms) * ADCval / (ADCnumValues - ADCval): */
	uint32_t divisor_shifted = (ADC_MAX - val) >> 6;
	/* Rcalc is 4*Ra */
	uint32_t R = (Rcal * val); DIVROUND(R, divisor_shifted); /* R factor: 256 */

  /* Supported R range: up to 91752 */
	int32_t T;
	uint32_t Tu;
  if(R < 100*256) {
    Tu = 4292218517UL; /* T factor: 100 * 2^34; coefficient: +2.4984e-3 * R^2 */
    Tu -= (38051UL * R); /* T factor: 100 * 2^8 * 2^26; coefficient: -5.67e-6 *   R^3 */
    DIVROUND(Tu, 65536UL); /* T factor: 100 * 2^18 */
    Tu *= R; /* T factor: 100 * 2^26 */

    DIVROUND(Tu, 8); /* T factor: 100 * 2^23 */
    Tu += 1868679872UL; /* T factor: 100 * 2^23; coefficient:  2.22764 * R */
    DIVROUND(Tu, 131072UL); /* T factor: 100 * 2^6 */
    Tu *= R; /* T factor: 100 * 2^14 */

		/* XXX: Optimise away signed division? */
    T = Tu;
    T -= 396620595L; /* T factor: 100 * 2^14; coefficient: -242.078 */
    DIVROUND(T, 16384L); /* T factor: 100 */
  } else {
    Tu = 35942UL * R ; /* T factor: 100 * 2^38 * 2^8; coefficient: 1.30755e-    09 * R^4 */

    Tu += 549920477UL; /* T factor: 100 * 2^46; coefficient: 7.81484e-08 *      R^3 */
    DIVROUND(Tu, 131072UL); /* T factor: 100 * 2^29 */
    Tu *= R; /* T factor: 100 * 2^37 */

    DIVROUND(Tu, 8); /* T factur: 100 * 2^34 */
    Tu += 1504836281UL; /* T factor: 100 * 2^34; coefficient: 0.00087593 * R^2 */
    DIVROUND(Tu, 131072UL); /* T factor: 100 * 2^17 */
    Tu *= R; /* T factor: 100 * 2^25 */

    DIVROUND(Tu, 4); /* T factor: 100 * 2^23 */
    Tu += 1992806105UL; /* T factor: 100 * 2^23; coefficient: 2.37561 * R */
    DIVROUND(Tu, 65536UL); /* T factor: 100 * 2^7 */
    Tu *= R; /* T factor: 100 * 2^15 */

    Tu /= 2; /* T factor: 100 * 2^14 */
    Tu -= 403909837UL;
    DIVROUND(Tu, 16384UL); /* T factor: 100 */
		T = Tu;
  }

	return T;
}

/*
 * get_Rcal:
 * Return the calibration value for Ra, the fixed voltage dividing resistor.
 */
static uint16_t get_Rcal(void) {
#define Ra_avgcount 4
	uint8_t i = Ra_avgcount + 1 /* Easier looping */;
	uint32_t ADCacc = 0;
	while(--i) ADCacc += read_adc();

	/* Pt100 has 100 Ohms at 0 °C, Rcal = Rcal_multiplicator * Ra: */
	uint32_t accu = Ra_avgcount * Rcal_multiplicator * 100 * ADC_MAX;
	accu -= (Ra_avgcount * 4 * 100) * (ADCacc / Ra_avgcount);
	DIVROUND(accu, ADCacc);

	return accu;
}

static void print_temp(uint8_t digits) {
	uint32_t val;
	if(temp < 0) {
		lcd_buf[0] = '-';
		val = -temp;
	} else {
		lcd_buf[0] = ' ';
		val = temp;
	}
	char* start = lcd_buf + 1;

	if((int32_t)temp < 0) temp = -(int32_t)temp;


	uint8_t i = digits;
	do {
		*start++ = ' ';
	} while(--i);
	*++start = '\0';

	/* Output the digits from right to left: */
	do {
		/* Adjust the digit n to an ASCII value: */
		*--start = (val % 10) + '0';
		val /= 10;
	} while(val);
}

/*
 * check_cal:
 * Check the calibration jumper and guide the user through the
 * calibration process.
 */
static void check_cal(void) {
	/* Set pull-up for calibration jumper: */
	SET_PULLUP(CAL_PIN); _delay_ms(1);
	/*
	 * Check if the calibration jumper is inserted (closed) or
	 * if the calibration resistor value is blank (erased EEPROM).
	 * If so, enter calibration routine.
	 */
	if(!IS_SET(CAL_PIN) || Rcal == 0xFFFF) {
		lcd_home();
		lcd_putS(PSTR("?? @ 0" DEGREE "C"));
		lcd_gotoxy(0,1);
		lcd_putS(PSTR("Ra WAIT "));

		uint8_t ok = 0;
		const char* msg;
		do {
			strcpy_P(lcd_buf, PSTR("        "));
			Rcal = get_Rcal();
			temp = Rcal;

			ok = 0;
			msg = PSTR("FAIL@0" DEGREE "C");
			lcd_clear();
			lcd_home();
			/* Rcal must be within 10% range: */
			if(Rcal > (uint16_t)(0.9 * Ra * Rcal_multiplicator) &&
					Rcal < (uint16_t)(1.1 * Ra * Rcal_multiplicator)) {
				ok = 1;
				msg = PSTR("OK @ 0" DEGREE "C");
				lcd_putS(msg);
				temp *= 100 / Rcal_multiplicator;
				print_temp(5);
				lcd_buf[7] = lcd_buf[6];
				lcd_buf[6] = lcd_buf[5];
				lcd_buf[5] = ',';
			} else {
				lcd_putS(msg);
				temp /= Rcal_multiplicator;
				print_temp(5);
			}
			/* Insert decimal separator: */
			lcd_gotoxy(0,1);
			lcd_puts(lcd_buf + 1);
			lcd_putc('\xF4'); /* \Omega in charset */

		} while(!IS_SET(CAL_PIN));

		/* Only update the EEPROM value if the update succeeded: */
		if(ok)
			eeprom_update_word((void*) EEP_RA_ADDR, Rcal);
	}
	/* Clear pullup, unneeded: */
	RESET(CAL_PIN);
}
